/**
   We need to do three things to each model.

   1) Make a copy of this model so that future pipeline
   stages can mutate the copy as they render it. In
   addition, traverse this model's DAG of nested models
   and recursively copy them. This will convert the DAG
   (directed acyclic graph) of nested models into a tree
   of nested models. For example, if this model has the
   following DAG structure,
               m1
              /  \
            /      \
           m2       m3
            \      /
              \  /
               m4
   then after the first step of this stage we will have
   the following tree structure.
               m1'
              /  \
            /      \
          m2'       m3'
          |         |
          |         |
         m4'        m4"
   where m1', m2', m3' are copies of m1, m2, m3, and
   m4' and m4" are both copies of m4.

   2) Transform all the vertices (and their normals)
   in the nested models into a single model coordinate
   system.

   3) Flatten the triangle lists from the nested models
   into one single triangle list in the top level model.
   (We only need to do this if we want to keep the rest
   of the pipeline the same as it was before we added
   nested models. We could instead rewrite the other
   stages to traverse the tree of nested models.)
*/
import java.util.LinkedList;

public class P0_FlattenModel implements PipelineStage
{
   public P0_FlattenModel()
   {
   }

   /**
      Copy, transform to model coordinates, and flatten
      all of the sub-models of this model.
   */
   public Model processModel(Model model)
   {
      // Create a new Model that is a copy of the original Model.
      // This copied model has the same nested model structure as
      // the original model, but it is a Tree instead of a DAG.
      Model newModel = copy( model );

      // For each sub-model of the new Model, transform all of its
      // vertices to one model coordinate system. This transformed
      // model still has the same nested model structure as the
      // original model but it has all of its triangles in the same
      // model coordinate system.
      for (Model nestedModel : newModel.nestedModels)
      {
         transform( nestedModel, new Matrix(1.0) );
      }

      // For each sub-model of the new Model, copy all of its
      // triangles into the triangle list of the top level model.
      for (Model nestedModel : newModel.nestedModels)
      {
         flatten( nestedModel, newModel.triangleList );
      }

      return newModel;
   }


   /**
      This is essentially a pre-order traversal of the tree of nested models.
   */
   public Model copy(Model model)
   {
      Model newModel = new Model(model); // copy everything except the geometry

      // Copy all the triangles in the model's triangle list.
      // (This can be a shallow copy because in the next method
      // we will make deep copies of the Triangle's vertices.)
      for (Triangle t : model.triangleList)
      {
         Triangle tCopy = new Triangle(t);
         newModel.addTriangle( tCopy );
      }

      // Recursively copy all the triangles in the sub-models.
      for (Model nestedModel : model.nestedModels)
      {
         newModel.addModel( copy(nestedModel) );
      }

      return newModel;
   }//copy()


   /**
      This is also essentially a pre-order traversal of the tree of nested models.
   */
   public void transform(Model model, Matrix ctm)
   {
      ctm = ctm.times( model.modelMatrix );

      Matrix invTranspose = ctm.invTranspose();

      // Transform all the triangles in the model's triangle list.
      for (Triangle t : model.triangleList)
      {
         // transform vertices
         t.v[0] = ctm.times(t.v[0]);
         t.v[1] = ctm.times(t.v[1]);
         t.v[2] = ctm.times(t.v[2]);
         // transform normal vectors
         if (null != t.v[0].n) t.v[0].n = invTranspose.times(t.v[0].n);
         if (null != t.v[1].n) t.v[1].n = invTranspose.times(t.v[1].n);
         if (null != t.v[2].n) t.v[2].n = invTranspose.times(t.v[2].n);
      }

      // Recursively transform all the triangles in the sub-models.
      for (Model nestedModel : model.nestedModels)
      {
         transform(nestedModel, ctm);
      }
   }//transform()


   /**
      This is also a pre-order traversal of the tree of nested models.
   */
   private void flatten(Model model, LinkedList<Triangle> triangleList)
   {
      // Flatten all the triangles in this model's triangle list.
      triangleList.addAll( model.triangleList );

      // Recursively flatten all the sub-models.
      for (Model nestedModel : model.nestedModels)
      {
         flatten(nestedModel, triangleList);
      }
   }//flatten()
}